#include <Arduino.h>

#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>

#include <ESP8266WebServer.h>

#include <ESP8266mDNS.h>
#include <ArduinoOTA.h>

#include <EEPROM.h>

#include <NTPClient.h>
#include <WiFiUdp.h>

#include <OneWire.h>
#include <DallasTemperature.h>

ESP8266WiFiMulti wifiMulti;
ESP8266WebServer server(80);

WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, "pool.ntp.org");

const char* ssid1     = "SSID1";
const char* password1 = "PASS";
const char* ssid2     = "SSID2";
const char* password2 = "PASS";

const byte led = 2;
bool keypress;
byte button;

const int rel[] = {D7, D6, D5, D2};

OneWire oneWire(D1);
DallasTemperature sensors(&oneWire);

const unsigned int vcnt = 4; // relays count - num_elements_of(relays[])
const unsigned int maxrules = 30; // esp can't send large html, so 48 rules is max for now without optimisations

bool state[vcnt];

// EEPROM partitioning: 0-3 - valves 0-3 state save; 4 - timezone; 5 - rules number; 6-36 - ssid length(1) + ssid(29); 37-68 - pass length(1) + pass(30)
// timer partitioning: X - enabled/disabled; X+1 - relay number (0-3); X+2 / X+3 - timer hours and minutes; X+4 - action for timer (Dis/En); X+5 - action for thermos (Dis/Less/More); X+6 - temp
// so 1st timer will be at 70(eeppref)-71-72-73-74-75-76 addresses (70+somevar*bw = 70+0*7); 2nd - eeppref+somevar*bw - 70+1*7 = 77-78-79-80-81-82; etc
const byte bw = 7;

const byte eeppref = 70;

const String relays[] = {"Зал / окна", "Зал / лестница", "Кухня", "Канал 4"};
const String actions[] = {"Выключить", "Включить"};
const String termos[] = {"Без термостата", "Если ниже", "Если выше"};

void setup() {
//  Serial.begin(115200);
//    Serial.setDebugOutput(true);
  EEPROM.begin(eeppref);

  int rules = EEPROM.read(5);
  if (rules > maxrules) {
    rules = maxrules; // rules limit
  }

  EEPROM.end();
  EEPROM.begin(rules*bw+eeppref);

  String ssid3;
  for (int s = 7; s < EEPROM.read(6)+7; s++) {
  ssid3 += char(EEPROM.read(s));
  }

  String password3;
  for (int p = 38; p < EEPROM.read(37)+38; p++) {
    password3 += char(EEPROM.read(p));
  }
  
  pinMode(led, OUTPUT);
  for (int r = 0; r < vcnt; r++)
  {
  pinMode(rel[r], OUTPUT);
  digitalWrite(rel[r], HIGH);
  }

// wifi connection section
  WiFi.mode(WIFI_OFF);        //Prevents reconnection issue (taking too long to connect)
  delay(1000);
  WiFi.mode(WIFI_STA);        //This line hides the viewing of ESP as wifi hotspot
  // WiFi.begin(ssid, password);     //Connect to your WiFi router
  wifiMulti.addAP(ssid1, password1);
  wifiMulti.addAP(ssid2, password2);
  // https://forum.arduino.cc/t/string-to-char-solved/419774
  wifiMulti.addAP(ssid3.c_str(), password3.c_str());
 
// Wait for connection
  while (wifiMulti.run() != WL_CONNECTED) {
    digitalWrite(led, HIGH);
    delay(250);
    digitalWrite(led, LOW);
    delay(250);
  }

  int timezone = EEPROM.read(4);
  if (timezone > 24) {
    timezone = 12; // if EEPROM value not set
  }  
  timeClient.begin();
  timeClient.setTimeOffset((timezone-12)*3600);
  timeClient.setUpdateInterval(60000);

// OTA section
  // Port defaults to 8266
  // ArduinoOTA.setPort(8266);

  // Hostname defaults to esp8266-[ChipID]
  // ArduinoOTA.setHostname("pzem_meter");
  char tmp[15];
  sprintf(tmp, "floor-%06x", ESP.getChipId());
  ArduinoOTA.setHostname(tmp);

  // No authentication by default
  ArduinoOTA.setPassword("temp123");

  // Password can be set with it's md5 value as well
  // MD5(admin) = 21232f297a57a5a743894a0e4a801fc3
  // ArduinoOTA.setPasswordHash("21232f297a57a5a743894a0e4a801fc3");

  ArduinoOTA.begin();

  sensors.begin(); // ds18b20
  sensors.setResolution(10);
  
// HTTP server setup
  server.on("/", handle_OnConnect);
  
  server.on("/rel0on", handle_rel0on);
  server.on("/rel0off", handle_rel0off);

  server.on("/rel1on", handle_rel1on);
  server.on("/rel1off", handle_rel1off);
  
  server.on("/rel2on", handle_rel2on);
  server.on("/rel2off", handle_rel2off);
  
  server.on("/rel3on", handle_rel3on);
  server.on("/rel3off", handle_rel3off);
 
  server.on("/timer", handle_timer);
  server.on("/settings", handle_settings);
  server.on("/temp", handle_temp);

  server.onNotFound(handle_NotFound);

  server.begin();
  for (int e = 0; e < vcnt; e++) {

    if(EEPROM.read(e) == 1)
    {
      state[e] = HIGH;
    }
    else
    {
      state[e] = LOW;
    }
  }
}

void loop() {
  while (wifiMulti.run() != WL_CONNECTED) {
    digitalWrite(led, HIGH);
    delay(250);
    digitalWrite(led, LOW);
    delay(250);
  }
  
  delay(100);
  server.handleClient();
  ArduinoOTA.handle();
  timeClient.update();

  int rules = EEPROM.read(5);
  
  button = analogRead(A0)*5/1024;

  if (button == 0) {
    keypress = 0;
  } else {
    if (keypress != 1) {
     state[button-1] = !state[button-1];
     keypress = 1;
    }
  }
  
// where is isTimeSet()?!
  if(timeClient.getEpochTime() > 1600000000) {
    for (int t = 0; t < rules; t++) {
     if (EEPROM.read(eeppref+t*bw) == 1) { // check if timer enabled
      if (EEPROM.read(eeppref+t*bw+2) == timeClient.getHours()) { // check if hour match 
        if (EEPROM.read(eeppref+t*bw+3) == timeClient.getMinutes()) { // check if minute match
          if (timeClient.getSeconds() == 0) { // work only on 0th second
            int do_valve = EEPROM.read(eeppref+t*bw+1); // get valve number from EEPROM
            int do_action = EEPROM.read(eeppref+t*bw+4); // get action from EEPROM
            int do_termos = EEPROM.read(eeppref+t*bw+5); // get termos action from EEPROM
            int do_temp = EEPROM.read(eeppref+t*bw+6);

            int temperatureC;
            if (do_termos != 0) {
              sensors.requestTemperatures(); 
              int temperatureC = sensors.getTempCByIndex(0);
            }
              
              if (do_termos == 0 || (do_termos == 1 && temperatureC < do_temp-10) || (do_termos == 2 && temperatureC > do_temp-10)) {
                if (do_action == 0) { // if action "close"
                  state[do_valve] = 0; // close valve
                } else if (do_action == 1) { // if action "open"
                  state[do_valve] = 1; // open valve
                } 
              }
            
          }
        }
      }
     }
    }
   }

  for (int s = 0; s < vcnt; s++) {
    if(state[s]) {
      digitalWrite(rel[s], LOW);
         if (EEPROM.read(s) != 1)
         {
          EEPROM.write(s, 1);
          EEPROM.commit();
         }
     }
      else
     {
      digitalWrite(rel[s], HIGH);
       if (EEPROM.read(s) != 0)
        {
          EEPROM.write(s, 0);
          EEPROM.commit();
        }
      }
  } 

}
  
// print uptime
String millis2time(){
 String Time="";
 unsigned long ss;
 byte mm, hh;
 ss = millis()/1000;
 hh = ss/3600;
 mm = (ss-hh*3600)/60;
 ss = (ss-hh*3600)-mm*60;
 if (hh<10) Time+="0";
 Time += (String)hh + ":";
 if (mm<10) Time += "0";
 Time += (String)mm + ":";
 if (ss<10) Time += "0";
 Time += (String)ss;
 return Time;
}

void handle_NotFound() {
  server.send(404, "text/plain", "Not found");
}

void handle_OnConnect() {
  server.send(200, "text/html", SendHTML());
}

// Наверняка все, что ниже, можно было сделать красивее :)
// Например, сделать общую страницу и передавать номер кнопки, как параметр.
// Но не будем уже ломать то, что работает.
void handle_rel0on()
{
  state[0] = HIGH;
  server.send(200, "text/html", SendHTML());
}

void handle_rel0off()
{
  state[0] = LOW;
  server.send(200, "text/html", SendHTML());
}

void handle_rel1on()
{
  state[1] = HIGH;
  server.send(200, "text/html", SendHTML());
}

void handle_rel1off()
{
  state[1] = LOW;
  server.send(200, "text/html", SendHTML());
}

void handle_rel2on()
{
  state[2] = HIGH;
  server.send(200, "text/html", SendHTML());
}

void handle_rel2off()
{
  state[2] = LOW;
  server.send(200, "text/html", SendHTML());
}

void handle_rel3on()
{
  state[3] = HIGH;
  server.send(200, "text/html", SendHTML());
}

void handle_rel3off()
{
  state[3] = LOW;
  server.send(200, "text/html", SendHTML());
}

void handle_timer() {
  server.send(200, "text/html", showtimer());
}

void handle_settings() {
  server.send(200, "text/html", showsettings());
}

void handle_temp() {
  sensors.requestTemperatures(); 
  float temperatureC = sensors.getTempCByIndex(0);
  server.send(200, "text/html", String(temperatureC));
}

//String SendHTML(uint8_t serv0stat, uint8_t serv1stat, uint8_t serv2stat, uint8_t serv3stat, uint8_t serv4stat, uint8_t serv5stat, uint8_t serv6stat, uint8_t serv7stat)
String SendHTML()
{
  String ptr = "<!DOCTYPE html> <html>\n";
  ptr +="<head><meta name=\"viewport\" content=\"10;URL=/\" http-equiv=\"refresh\" charset=utf-8>\n";
  ptr +="<title>Контроллер теплого пола</title>\n";
  ptr +="<style>html { font-family: Helvetica; display: inline-block; margin: 0px auto; text-align: center;}\n";
  ptr +="body{margin-top: 50px;} h1 {color: #444444;margin: 50px auto 30px;} h3 {color: #444444;margin-bottom: 50px;}\n";
  ptr +=".button {display: block;width: 80px;background-color: #1abc9c;border: none;color: white;padding: 13px 30px;text-decoration: none;font-size: 25px;margin: 0px auto 35px;cursor: pointer;border-radius: 4px;}\n";
  ptr +=".button-on {background-color: #1abc9c;}\n";
  ptr +=".button-on:active {background-color: #16a085;}\n";
  ptr +=".button-off {background-color: #34495e;}\n";
  ptr +=".button-off:active {background-color: #2c3e50;}\n";
  ptr +="p {font-size: 14px;color: #888;margin-bottom: 10px;}\n";
  ptr +="</style>\n";
  ptr +="</head>\n";
  ptr +="<body>\n";
  ptr +="<h1>Контроллер теплого пола</h1>\n";
  ptr +="<h3>Подключен к точке: ";
  ptr += WiFi.SSID();
  ptr += "<small>. Время работы: ";
  ptr += millis2time();
  ptr += "</small><br>";
  ptr += "<a href=/timer>Таймер</a> <a href=/settings>Настройки</a> <a href=/temp>Температура</a></h3>\n";
  ptr += "<table border=0 width=50% align=center>";

  for (int i = 0; i < vcnt; i++) {
    ptr += "<tr><td align=left>";
    ptr += relays[i];
    ptr += "</td><td colspan=2><a class=\"button button-";
  
    if(state[i]) {
      ptr += "on\" href=\"/rel";
      ptr += i;
      ptr += "off\">On";
    }
    else
    {
      ptr += "off\" href=\"/rel";
      ptr += i;
      ptr += "on\">Off";
    }
  }

  ptr += "</table>\n</body>\n</html>\n";
  return ptr;
}

String showtimer() {

  String ptr = "<!DOCTYPE html> <html>\n";
  ptr +="<head><meta name=\"viewport\" charset=utf-8>\n";
  ptr +="<title>Hot Floor Control</title>\n";
  ptr +="<style>html { font-family: Helvetica; display: inline-block; margin: 0px auto; text-align: center;}\n";
  ptr +="body{margin-top: 50px;}\n";
  ptr +="p {font-size: 14px;color: #888;margin-bottom: 10px;}\n";
  ptr +="</style>\n";
  ptr +="</head>\n";
  ptr +="<body>\n";
  ptr += "<a href=/>Назад</a>. Текущее время: ";
  ptr += timeClient.getFormattedTime();
  ptr += "<br><br>\n";

  if (server.args() >= 4) {
    int cell = server.arg("cell").toInt();

    if (server.arg("on") == "on") {
      EEPROM.write(eeppref+cell, 1);
      }
      else
      {
      EEPROM.write(eeppref+cell, 0);
      }

      if (server.hasArg("valve")){
        EEPROM.write(eeppref+cell+1, server.arg("valve").toInt());
      }

      if (server.hasArg("action")){
        EEPROM.write(eeppref+cell+4, server.arg("action").toInt());
      }

      if (server.hasArg("termos")){
        EEPROM.write(eeppref+cell+5, server.arg("termos").toInt());
      }

      if (server.hasArg("temperature")){
        EEPROM.write(eeppref+cell+6, server.arg("temperature").toInt()+10);
      }
      
    int hours = server.arg("stime").substring(0,2).toInt(); // but where is ":"?
    int mins = server.arg("stime").substring(3).toInt();

    if (hours > 59) {
      hours = 59;
    }

    if (hours < 0) {
      hours = 0;
    }

    if (mins > 59) {
      mins = 59;
    }

    if (mins < 0) {
      mins = 0;
    }

    EEPROM.write(eeppref+cell+2, hours);
    EEPROM.write(eeppref+cell+3, mins);
    EEPROM.commit();
  }

int rules = EEPROM.read(5);
for (int i = 0; i < rules; i++) {
  ptr += "<form>\n<input type=checkbox name=on ";

   if (EEPROM.read(eeppref+i*bw) == 1) {
    ptr += "checked";
    }
   ptr += ">\n<input type=hidden name=cell value=";
   ptr += i*bw;
   ptr += ">\n<select name=valve>\n";
   for (int v = 0; v < vcnt; v++)
   {
    ptr += "<option value=";
    ptr += v;

    if (EEPROM.read(eeppref+i*bw+1) == v) {
      ptr += " selected";
    }
    ptr += ">";
    ptr += relays[v];
    ptr += "</option>\n";
   }
   ptr += "</select>\n";
   ptr += "Время:\n";
   ptr += "<input type=time name=stime required value='";
   if (EEPROM.read(eeppref+i*bw+2) < 10) {
    ptr += "0";
    }
   ptr += EEPROM.read(eeppref+i*bw+2);
   ptr += ":";
   if (EEPROM.read(eeppref+i*bw+3) < 10) {
    ptr += "0";
    }
   ptr += EEPROM.read(eeppref+i*bw+3);
   ptr += "'>";
   ptr += "</input>\n";
   ptr += "<select name=action>\n";
   
   for (int a = 0; a < 2; a++)
   {
    ptr += "<option value=";
    ptr += a;

    if (EEPROM.read(eeppref+i*bw+4) == a) {
      ptr += " selected";
    }
    ptr += ">";
    ptr += actions[a];
    ptr += "</option>\n";
   }
   
   ptr += "<br></select>\n";

   ptr += "<select name=termos>\n";
   for (int a = 0; a < 3; a++)
   {
    ptr += "<option value=";
    ptr += a;

    if (EEPROM.read(eeppref+i*bw+5) == a) {
      ptr += " selected";
    }
    ptr += ">";
    ptr += termos[a];
    ptr += "</option>\n";
   }
   
   ptr += "<br></select>\n";
/*
   ptr += "<select name=temperature>";
   for (int t = 0; t < 30; t++) {
    ptr += "<option value=";
    ptr += t;
    if (EEPROM.read(eeppref+i*bw+6) == t) {
      ptr += " selected";
    }
    ptr += ">";
    ptr += t-5;
    ptr += "</option>\n";
  }

  ptr += "</select>\n";
   */

   ptr += "<input type=number name=temperature min=-10 max=30 size=3 step=1 value=";
   ptr += EEPROM.read(eeppref+i*bw+6)-10;
   ptr += ">";


   
   ptr += "<input type=submit value=Сохранить>\n";
   ptr += "</form>\n\n";

}
  ptr += "</html>";
  return ptr;
}

String showsettings() {
    String ptr = "<!DOCTYPE html> <html>\n";
  ptr +="<head><meta name=\"viewport\" charset=utf-8>\n";
  ptr +="<title>Hot Floor Control</title>\n";
  ptr +="<style>html { font-family: Helvetica; display: inline-block; margin: 0px auto; text-align: center;}\n";
  ptr +="body{margin-top: 50px;}";
  ptr +="p {font-size: 14px;color: #888;margin-bottom: 10px;}\n";
  ptr +="</style>\n";
  ptr +="</head>\n";
  ptr +="<body>\n";
  ptr += "<a href=/>Назад</a>. Текущее время: ";
  ptr += timeClient.getFormattedTime();
  ptr += "<br><br>\n";
  int rules = EEPROM.read(5);
  
  if (server.hasArg("timezone")) {
    if (EEPROM.read(4) != server.arg("timezone").toInt()) {
    EEPROM.write(4, server.arg("timezone").toInt());
    EEPROM.commit();
    ESP.restart();
    }
  }

  if (server.hasArg("rulesnumber")) {
    EEPROM.write(5, server.arg("rulesnumber").toInt());
    EEPROM.commit();
  }

  if (server.hasArg("ssid") && server.hasArg("password")) {
    String ssid = server.arg("ssid");
    String password = server.arg("password");
    
    EEPROM.write(6, ssid.length());
    EEPROM.write(37, password.length());
    for (int i = 0; i < 30; i++) {
      EEPROM.write(7+i, ssid[i]);
      EEPROM.write(38+i, password[i]);
    }
    EEPROM.commit();
  }

  ptr += "<form>";
  ptr += "Часовой пояс: ";
  ptr += "<select name=timezone>";
  for (int i = 0; i < 25; i++) {
    ptr += "<option value=";
    ptr += i;
    if (EEPROM.read(4) == i) {
      ptr += " selected";
    }
    ptr += ">";
    ptr += i-12;
    ptr += "</option>\n";
  }

  ptr += "</select>\n<input type=submit value=Сохранить>\n";
  ptr += "</form>\n\n";

  ptr += "<p align=left><form>\n";
  ptr += "Количество таймеров: ";
  ptr += "<select name=rulesnumber>\n";
  for (int i = 0; i < maxrules; i++) {
    ptr += "<option value=";
    ptr += i;
    if (EEPROM.read(5) == i) {
      ptr += " selected";
    }
    ptr += ">";
    ptr += i;
    ptr += "</option>\n";
    
  }

  ptr += "</select>\n<input type=submit value=Сохранить>";
  ptr += "</form></p>\n\n";

  ptr += "<form>";
  ptr += "SSID: <input type=text maxlength=30 name=ssid value=";
  String ssid;
  for (int i = 7; i < EEPROM.read(6)+7; i++) {
    ssid += char(EEPROM.read(i));
  }
  ptr += ssid;
  
  ptr += ">\n<br>PASS: <input type=text maxlength=30 name=password value=";
  String password;
  for (int i = 38; i < EEPROM.read(37)+38; i++) {
    password += char(EEPROM.read(i));
  }
  ptr += password;
  
  ptr += ">\n<br><input type=submit value=Сохранить>";
  ptr += "</form>\n\n</html>";

  return ptr; 
}
